/***************************************************************************
 *   Copyright (C) 2004 by Christoph Freundl                               *
 *   Christoph.Freundl@informatik.uni-erlangen.de                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "textarea.h"
#include "fontfactory.h"

namespace SDLwidgets {

TextArea::TextArea( int width, int height ) {
	_width = width;
	_min_width = width;
	_height = height;
	_min_height = height;

	// default color for a text label is white
	_color.r = 255;
	_color.g = 255;
	_color.b = 255;
	
	_font = FontFactory::getInstance()->createFont( "tahoma.ttf", 12 );
}

TextArea::~TextArea() {
}

void TextArea::setFont( std::string fontName, int fontSize ) {
	_font = FontFactory::getInstance()->createFont( fontName, fontSize );
/*	renderText();*/
}

void TextArea::draw( SDL_Surface* surface ) {
	SDL_Rect srcRect, destRect;
	Sint16 offset_x, offset_y;
	getOffset( offset_x, offset_y );

	int maxHeight = TTF_FontHeight( _font );
	destRect.x = offset_x;
	destRect.y = offset_y + _height - maxHeight;
	srcRect.x = 0;
	srcRect.y = 0;
	srcRect.w = _width;
	srcRect.h = maxHeight;
	for ( std::vector<std::string>::reverse_iterator it = _line.rbegin();
	      it != _line.rend() && destRect.y >= offset_y; it++, destRect.y -= maxHeight + 1 ) {
		SDL_Surface* textSurface =
			TTF_RenderText_Solid( _font, it->c_str(), _color );
		SDL_BlitSurface( textSurface, &srcRect, surface, &destRect );
		SDL_FreeSurface( textSurface );
	}
}

}
