/* $Id: avatar.cpp,v 1.113 2005/05/10 07:51:42 pohlt Exp $ */

#include  "avatar.hpp"
#include  "serialize.hpp"
#include  "audio.hpp"
#include  "bonusshield.hpp"
#include  "bonusfuel.hpp"
#include  "bonushomingvirus.hpp"
#include  "rope.hpp"
#include  "hook.hpp"
#include  "smoke.hpp"

#include  <cmath>

/**********************************************************/

const real Avatar::INIT_AIMING_SPEED      =  1.0,
           Avatar::AIMING_SPEED_INCREMENT =  0.5,
           Avatar::MAX_AIMING_SPEED       = 10.0,
           Avatar::MAX_AIMING_ANGLE       = 180 - 15,
           Avatar::AIMING_ANGLE_TO_RAD = M_PI / 180.0;


const real Avatar::INIT_WALKING_SPEED        = 1.0,
           Avatar::MAX_WALKING_SPEED         = 5.0,
           Avatar::MAX_WALKING_SPEED_DIGGING = 2.0;

/**********************************************************/

Avatar::Avatar()
      : m_fuel( MAX_FUEL ),
        m_maxFuel( MAX_FUEL ),
        m_aimingAngle( 90.0 ),
        m_aimingSpeed( 0.0 ),
        m_walkingSpeed( INIT_WALKING_SPEED ),
        m_bonusPack( m_worldPointer ),
        m_Player( NULL )
{
	setState( FACING_RIGHT, true );
	setState( LIVING, false );
	setState( DIGGING, false );
	setState( JUMPING,false  );
	setState( MOVING, false );

	m_popupSample = Audio::getInstance()->loadSound( "sound/effects/popup.wav" );
	
	m_messageSink = &m_defaultMessageSink;
}

/**********************************************************/

Avatar::~Avatar() {
	DBG( 1 ) CHECK( m_Player == NULL,
	                "Avatar::~Avatar: m_Player != NULL\n" );
}

/**********************************************************/

void Avatar::setDirection( const Uint32 direction )
{
	DBG( 2 ) ASSERT( direction < NUM_DIRECTIONS,
	                 "Avatar::setDirection: passed direction %d "
	                 "is out of range\n" );

	// do not set the new direction, if it were the old one
	const Uint16 oldDirection = (isFacingRight() ? RIGHT : LEFT);
	if( direction == oldDirection ) return;
	setState( FACING_RIGHT, direction == RIGHT );
	m_walkingSpeed = INIT_WALKING_SPEED;
	updateSpriteSelection();
}

/**********************************************************/

void Avatar::setAimingDirection( const int dir ) {
	// if not turning set aiming speed to zero
	if( dir == 0 ) {
		m_aimingSpeed = 0.0;
	}
	else

	// if turning in positive direction
	if( dir > 0 ) {
		// if just started turning
		if( m_aimingSpeed < INIT_AIMING_SPEED ) {
			m_aimingSpeed = INIT_AIMING_SPEED;
		}
		else if( m_aimingSpeed < MAX_AIMING_SPEED ) {
			m_aimingSpeed += AIMING_SPEED_INCREMENT;
		}
		setAimingAngle( m_aimingAngle + m_aimingSpeed );
	}
	else
	// if turning in negative direction
	if( dir < 0 ) {
		// if just started turning
		if( m_aimingSpeed > -INIT_AIMING_SPEED ) {
			m_aimingSpeed = -INIT_AIMING_SPEED;
		}
		else if( m_aimingSpeed > -MAX_AIMING_SPEED ) {
			m_aimingSpeed -= AIMING_SPEED_INCREMENT;
		}
		setAimingAngle( m_aimingAngle + m_aimingSpeed );
	}
}

/**********************************************************/

void Avatar::dig() {
	m_worldPointer->getMap()->makeHole( ROUND( m_positionX + 8.0*getAimingDX() ),
		ROUND( m_positionY + 8.0*getAimingDY() ) - m_collRectY - m_collRectHeight - 2,
		m_collRectHeight/2 + 3, Flags::UNDIGGABLE | Flags::INDESTRUCTIBLE );
}

/**********************************************************/

void Avatar::testGrounded( bool& groundedMid, bool& groundedFW,
                           bool& groundedBW ) {
	int dummy;
	const int fwX = (isFacingRight() ? m_collRectX + m_collRectWidth : m_collRectX - 1),
	          bwX = (isFacingRight() ? m_collRectX - 1 : m_collRectX + m_collRectWidth);

	groundedMid = ! (testLinePassableRel( m_collRectX, m_collRectY + m_collRectHeight,
	                                      m_collRectX + m_collRectWidth - 1, m_collRectY + m_collRectHeight,
	                                      dummy, dummy, dummy, dummy ) &&
	                 testPassableRel( fwX, m_collRectY + m_collRectHeight )),
	groundedFW  = ! testLinePassableRel( fwX, m_collRectY,
	                                     fwX, m_collRectY + m_collRectHeight - 1,
	                                     dummy, dummy, dummy, dummy ),
	groundedBW  = ! testPassableRel( bwX, m_collRectY + m_collRectHeight );
}

/**********************************************************/

int Avatar::applyDamage( const Particles::ParticleData& p )
{
	// recoil by particles
	// factor 2 is unphysical but also more fun
	m_speedX += 2 * p.vel.x / m_mass;
	m_speedY += 2 * p.vel.y / m_mass;
	setState( GROUNDED, false );

	if( m_bonusPack.hasBonus(BONUS_SHIELD) ) {
		((BonusShield*)m_bonusPack[BONUS_SHIELD])->absorbDamage( p.vel );
		return 0;
	}

	const int damage =
	min( ROUND( p.damage * (p.vel - getSpeed()).abs2() ), m_health );

	LOG(5) INFO( "CollidableObject::applyDamage: health: "
	             "%i, damage: %i\n", m_health, damage );

	m_health -= damage;
	return damage;
}

/**********************************************************/

int Avatar::applyDamage( const int damage, Player* shooter )
{
	if( m_bonusPack.hasBonus(BONUS_SHIELD) )  return 0;

	return CollidableObject::applyDamage( damage, shooter );
}

/**********************************************************/

void Avatar::pickUp( Bonus* bonus )
{
	// The collission rectangle of the bonus must not be set
	// at this stage, since it will never be removed after
	// being picked up by an avatar.
	m_worldPointer->unhookObject( bonus );

	// let the bonus have "the last word" while being picked up
	bonus->beingPickedUp( this );
	
	// apply the bonus to the avatar
	bonus->apply( this );
	// if necessary, pack the bonus into the bonus pack
	if( bonus->mustBePutIntoPack() ) {
		// detach all attached objects from the bonus, because the
		// bonus object itself lives furtheron, in the bonus pack
		bonus->detachFromAll();
		m_bonusPack.add( bonus );
	// otherwise delete the consumed bonus
	} else {
		delete bonus;
	}
}

/**********************************************************/

void Avatar::getLockedByHomingMissile( HomingMissile *const missile,
                                       const real distanceSquare   )
{
	// If the homing missile is inside the minimum transfer distance
	// for the homing missile virus, AND this avatar has got this bonus,
	// apply the virus bonus to the missile.
	if(    distanceSquare <= BonusHomingVirus::getInfectionRadiusSquare()
	    && getBonusPack().hasBonus(BONUS_HOMING_VIRUS) ) {
		(static_cast<BonusHomingVirus*>
			(getBonusPack()[BONUS_HOMING_VIRUS]))->applyVirus( missile, this );
	}
}

/**********************************************************/

void Avatar::update() {
	if( m_damageSampleTimeout > 0 ) m_damageSampleTimeout--;

	Map* const map = m_worldPointer->getMap();

	if ( !isLiving() ) {
		return;
	}
	
	const real DT = 1.0;

	int cx, cy, pcx, pcy, dummy;

	// remove avatar stencil
	removeCollRect();

	if( ! isMoving() ) m_walkingSpeed = INIT_WALKING_SPEED;
	else m_walkingSpeed += 0.05;

	if( isDigging() ) dig();

	if( isMoving() ) {
		// increment the animation frame
		if( ++m_Frame >= m_SequenceLength ) m_Frame = 0;
	}

	bool groundedMid, groundedFW, groundedBW;
	testGrounded( groundedMid, groundedFW, groundedBW );

	LOG( 5 ) INFO( "grounded status A: %c%c%c\n",
	               (groundedBW  ? '#' : '.'),
	               (groundedMid ? '#' : '.'),
	               (groundedFW  ? '#' : '.') );

	if( isGrounded() && ( ! (groundedMid || groundedFW || groundedBW) ) ||
	                       getSpeed().abs2() > 36.0) {
		LOG( 5 ) INFO( "Avatar::update: GROUNDED lost A\n" );
		setState( GROUNDED, false );
	}

	if( isGrounded() ) {
		if( isMoving() ) {

			if( isDigging() ) {
				if( m_walkingSpeed > MAX_WALKING_SPEED_DIGGING ) m_walkingSpeed = MAX_WALKING_SPEED_DIGGING;
			}
			else {
				if( m_walkingSpeed > MAX_WALKING_SPEED ) m_walkingSpeed = MAX_WALKING_SPEED;
			}

			real remainingWalkingDistance = m_walkingSpeed * DT;
			real currentSpeedX = 0, currentSpeedY = 0;

			while( isGrounded()  && (remainingWalkingDistance > 0.0) ) {

				LOG( 5 )
				INFO( "mid: %3i, %3i\n%c%c%c\n%c%c%c\n%c%c%c\n",
				      ROUND( m_positionX ), ROUND( m_positionY ), 
				      (testPassableRel( -1, -1 ) ? '.' : '#'),
				      (testPassableRel(  0, -1 ) ? '.' : '#'),
				      (testPassableRel(  1, -1 ) ? '.' : '#'),
				      (testPassableRel( -1,  0 ) ? '.' : '#'),
				      (testPassableRel(  0,  0 ) ? '.' : '#'),
				      (testPassableRel(  1,  0 ) ? '.' : '#'),
				      (testPassableRel( -1,  1 ) ? '.' : '#'),
				      (testPassableRel(  0,  1 ) ? '.' : '#'),
				      (testPassableRel(  1,  1 ) ? '.' : '#') );

				m_speedX = m_speedY = 0;

				if( isDigging() ) dig();

				if( groundedFW && ! ( ! groundedMid && groundedBW) ) {
					m_speedY -= (remainingWalkingDistance >= 1.0 ? 1.0 : remainingWalkingDistance);
					LOG( 5 ) INFO( "Avatar::update: walking up\n" );
				}
				else
				if ( groundedMid ) {
					m_speedX += (isFacingRight() ? 1 : -1) *
					            (remainingWalkingDistance >= 1.0 ? 1.0 : remainingWalkingDistance);
					LOG( 5 ) INFO( "Avatar::update: walking horiz.\n" );
				}
				else {
					m_speedY += (remainingWalkingDistance >= 1.0 ? 1.0 : remainingWalkingDistance);
					LOG( 5 ) INFO( "Avatar::update: walking down\n" );
				}

				const real newPosX = m_positionX + m_speedX,
				            newPosY = m_positionY + m_speedY;

				if( map->testCollRect( ROUND( newPosX ) + m_collRectX,
				                       ROUND( newPosY ) + m_collRectY,
				                       m_collRectWidth, m_collRectHeight, dummy, dummy ) ) {
					m_positionX = newPosX;
					m_positionY = newPosY;
					currentSpeedX += m_speedX;
					currentSpeedY += m_speedY;
				}

				remainingWalkingDistance--;

				// still standing on ground?
				testGrounded( groundedMid, groundedFW, groundedBW );

				LOG( 5 ) INFO( "grounded status B: %c%c%c\n",
				               (groundedBW  ? '#' : '.'),
				               (groundedMid ? '#' : '.'),
				               (groundedFW  ? '#' : '.') );

				if( ! (groundedMid || groundedFW || groundedBW) ) {
					// no, so remove flag
					LOG( 5 ) INFO( "Avatar::update: GROUNDED lost B\n" );
					setState( GROUNDED, false );
				}
			}
			m_speedX = currentSpeedX;
			m_speedY = currentSpeedY;
		}
		else {
			m_speedX = m_speedY = 0;
		}

		if( isJumping() ) {
			LOG( 5 ) INFO( "Avatar::update: jumping => GROUNDED lost\n" );
			m_speedY -= DT * (GRAVITATION + 4.5);
			setState( GROUNDED, false );
		}
	}

	else { // not standing on ground
		LOG( 5 ) INFO( "Avatar::update: not GROUNDED\n" );

		if( isMoving() ) {
			m_speedX += (isFacingRight() ? 0.1 : -0.1);
		}
		if( isJumping() ) {
			if( getFuel() > 0 ) {
				for( int i = 0; i < 15; i++ ) {
					m_worldPointer->getParticles()->addParticle(
					    Particles::ParticleData(
					    Vector( m_positionX, m_positionY + m_collRectY + m_collRectHeight ),
					    Vector( 5.0*(0.5 - globalRnd.getNormedReal()) + (isFacingRight() ? -2 : 2),
					    5.0*(1.0 + globalRnd.getNormedReal()) ),
					    Particles::BOUNCE_EARTH | //Particles::BOUNCE_OBJECT |
					    Particles::REDUNDANT | Particles::SPARK, 2, 0,
					    Particles::m_N_SPARK_COLORS/2 + globalRnd.getUint32() % (Particles::m_N_SPARK_COLORS/2+1), 1,
					    0xffffff ) );
				}
				m_speedX += 0.08 * (isFacingRight() ? 1 : -1);
				m_speedY -= DT * (GRAVITATION + 0.2);
				m_fuel--;
			}
			else {
				if( getFuel() == 0 ) {
					// smoke particles
					Uint32 smokeColor;
					for( int i = 0; i < 15; i++ ) {
						smokeColor = globalRnd.getUint32Between( 150, 240 );
						smokeColor += (smokeColor << 8) + (smokeColor << 16);
						m_worldPointer->getParticles()->addParticle(
						    Particles::ParticleData(
						    Vector( m_positionX, m_positionY + m_collRectY + m_collRectHeight ),
						    Vector( 4.0*(0.5 - globalRnd.getNormedReal()) + (isFacingRight() ? -2 : 2),
						    4.0*(1.0 + globalRnd.getNormedReal()) ),
						    Particles::BOUNCE_EARTH | //Particles::BOUNCE_OBJECT |
						    Particles::REDUNDANT, 0, 0,
						    25, 1,
						    smokeColor ) );
					}
					// smoke sprites
					Uint32 nSmokeClouds = (globalRnd.getUint32() & 1) + 1;
					while( nSmokeClouds-- ) {
						Smoke *smoke = static_cast<Smoke*>(
							m_worldPointer->newObject( (globalRnd.getUint32() & 3)
							                           ? SMOKE_15 : SMOKE_30) );
						const Vector smokeOffset( localRnd.getSint32Between( 0, 4 ) *
						                          (isFacingRight() ? -1 : 1),
						                          localRnd.getUint32() & 7 );
						smoke->setPosition( getPosition() + smokeOffset );
/*
						smoke->setPosition( m_positionX + (isFacingRight()
						                     ? localRnd.getSint32Between(-4,0)
						                     : localRnd.getSint32Between( 0,4)),
						                    m_positionY + m_collRectY
						                     + m_collRectHeight
						                     + (localRnd.getUint32() & 7) );
*/
						smoke->setSpeed( getSpeed() + localRnd.getVector( 2.0 ));
						smoke->setColor( 0 ); //TPLCOL_RGB(100,100,100) );
					}
					// new value for fuel
					m_fuel = globalRnd.getSint32Between( -8, 9 );
				}
				else {
					m_fuel++;
				}
			}

			/*
			// used to be WimPack
			else {
				m_speedX += 0.01 * (isFacingRight() ? 1 : -1);
				m_speedY -= DT * (GRAVITATION - 0.05);
			}
			*/
			
		}
		m_speedY += GRAVITATION*DT;

		const real newPosX = m_positionX+m_speedX*DT,
		           newPosY = m_positionY+m_speedY*DT;

		LOG( 5 ) INFO( "AAAAA: %.9e %.9e\n", newPosX, newPosY );

		// is there something to collide?
		if( map->testLineCollRect( ROUND( m_positionX ) + m_collRectX,
		                           ROUND( m_positionY ) + m_collRectY,
		                           ROUND( newPosX     ) + m_collRectX,
		                           ROUND( newPosY     ) + m_collRectY,
		                           m_collRectWidth, m_collRectHeight,
		                           pcx, pcy, cx, cy ) ) {
			// no, then just keep on flying
			m_positionX = newPosX;
			m_positionY = newPosY;
		}

		// yes, there is an in-flight collision
		else {
			LOG( 5 ) INFO( "Avatar::update: collision when moving from (%f, %f) to (%f, %f)\n",
			               m_positionX, m_positionY, newPosX, newPosY );
			m_positionX = pcx - m_collRectX;
			m_positionY = pcy - m_collRectY;

			DBG( 4 ) {
				CHECK( map->getFlags()->getValueAt( cx, cy ) & (Flags::OBSTACLE_EARTH | Flags::OBSTACLE_OBJECT),
				       "Avatar::update: (cx, cy) is passable %4i, %4i\n", cx, cy );
				CHECK( ! (map->getFlags()->getValueAt( pcx, pcy ) & (Flags::OBSTACLE_EARTH | Flags::OBSTACLE_OBJECT)),
				       "Avatar::update: (pcx, pcy) is not passable %4i, %4i\n", pcx, pcy );
				CHECK( ! (map->getFlags()->getValueAt( ROUND( m_positionX ),
				                                    ROUND( m_positionY ) ) & (Flags::OBSTACLE_EARTH | Flags::OBSTACLE_OBJECT)),
				       "Avatar::update: (posX, posY) is not passable %f, %f\n", m_positionX, m_positionY );
			}
			
			LOG( 5 ) INFO( "Avatar::update: setting pos after collision to (%f, %f)\n",
			               m_positionX, m_positionY );

			DBG( 3 ) CHECK( map->testCollRect( ROUND( m_positionX ) + m_collRectX,
			                                   ROUND( m_positionY ) + m_collRectY,
			                                   m_collRectWidth, m_collRectHeight, dummy, dummy ),
			                "Avatar::update: Avatar::update: setting to invalid position\n" );

			testGrounded( groundedMid, groundedFW, groundedBW );

//##############################################################################
/*
// groundMid
map->drawLine( ROUND( m_positionX + m_collRectX ), ROUND( m_positionY + 1 ),
               ROUND( m_positionX + m_collRectX + m_collRectWidth - 1 ), ROUND( m_positionY + 1 ), 0xff0000 );
map->setColorAt( ROUND( m_positionX + fwX ),
                 ROUND( m_positionY + 1 ), 0xff0000 );
// groundFW
map->drawLine( ROUND( m_positionX + fwX ), ROUND( m_positionY + m_collRectY ),
               ROUND( m_positionX + fwX ), ROUND( m_positionY + m_collRectY + m_collRectHeight - 1 ), 0x0000ff );
// groundBW
map->setColorAt( ROUND( m_positionX + bwX ),
                 ROUND( m_positionY + 1 ), 0x00ff00 );
*/
//##############################################################################

			const real speed2 = getSpeed().abs2();
			if( speed2 < 25.0 && (groundedMid /*|| groundedFW || groundedBW*/) ) {
				LOG( 5 ) INFO( "Avatar::update: GROUNDED again\n" );
				setState( GROUNDED, true );
				m_speedX = m_speedY = 0;
			}
			else {
				const real speed = SQRT_REAL( speed2 ),
				             outAngle = map->calculateBounceAngle( m_speedX, m_speedY, cx, cy, Flags::OBSTACLE_EARTH | Flags::OBSTACLE_OBJECT );
				m_speedX = speed * (speed > 1.5 ? 0.5 : 1.0) * COS_REAL( outAngle );
				m_speedY = speed * (speed > 1.5 ? 0.5 : 1.0) * SIN_REAL( outAngle );
			}
		}
	}

	LOG( 5 ) INFO( "BBBBB: %.9e %.9e\n", m_positionX, m_positionY );

	DBG( 5 ) map->setColorAt( ROUND( m_positionX ),
	                          ROUND( m_positionY ), 0xffffff );
	// generate blood
	for( int i = 0; i < 5 - ((10*m_health)/m_maxHealth); i++ ) {
		const real brightness = 0.7 + 0.3*globalRnd.getNormedReal();
		const Uint32 skincolor = getSkinColor();
		Uint32 r = ROUND( static_cast<Uint8>( skincolor >> 16 ) * brightness );
		Uint32 g = ROUND( static_cast<Uint8>( skincolor >>  8 ) * brightness );
		Uint32 b = ROUND( static_cast<Uint8>( skincolor       ) * brightness );

		m_worldPointer->getParticles()->addParticle( Particles::ParticleData(
		     Vector( m_positionX + m_collRectX + m_collRectWidth/2,
		             m_positionY + m_collRectY + m_collRectHeight/2 ),
		     Vector( 3.0*(0.5 - globalRnd.getNormedReal()),
		             3.0*(0.5 - globalRnd.getNormedReal()) ) + getSpeed(),
		     Particles::BOUNCE_EARTH | Particles::STAIN | Particles::REDUNDANT, 0,
		     0, 75, 1, (((r << 8) | g) << 8) | b ));
		}

	// shoot normal weapons
	if( isShooting() ) {
		getPlayer()->getActiveWeapon()->shoot( m_worldPointer, this );
	}

	// handle rope/hook actions
	
	// cut rope
	if( getState( ROPE_OFF )) {
		getPlayer()->getWeapon( Player::m_MAX_WEAPONS-1 )->shoot( m_worldPointer, this );
	}

	// release rope
	if( getState( ROPE_REL )) {
		getPlayer()->getWeapon( Player::m_MAX_WEAPONS-2 )->shoot( m_worldPointer, this );
	}

	// shoot rope
	if( getState( ROPE_ON )) {
		getPlayer()->getWeapon( Player::m_MAX_WEAPONS-3 )->shoot( m_worldPointer, this );
	}

	// update bonus pack
	m_bonusPack.update( this );

	// set avatar stencil
	LOG( 5 ) INFO( "CCCCC: %.9e %.9e\n", m_positionX, m_positionY );
	setCollRect();
}

/**********************************************************/

void Avatar::spawnAvatar() {
	Map* const map = m_worldPointer->getMap();
	int x = 0, y = 0, dummy;

	// reset avatar velocity
	m_speedX = m_speedY = 0;

	// reset weapons
	getPlayer()->rechargeWeaponsNow();
	
	while( true ) {
		x = m_collRectWidth + (globalRnd.getUint32() % (map->getSizeX()-2*m_collRectWidth)),
		y = m_collRectHeight + (globalRnd.getUint32() % (map->getSizeY()-2*m_collRectHeight));

		map->makeHole( x, y, m_collRectWidth + m_collRectHeight, Flags::UNDIGGABLE );

		if( map->testFilledCollRect( x + m_collRectX, y + m_collRectY,
		                             m_collRectWidth, m_collRectHeight, dummy, dummy ) )
			break;
	}

	// add initial shield bonus
	BonusShield* shield = dynamic_cast<BonusShield*>(
		m_worldPointer->newObject( BONUS_SHIELD ));
	shield->setProtectionTime( 100 ); // 100 frames = 4 seconds
	pickUp( shield );

	// play sound
	Audio::getInstance()->playSound( m_popupSample, ROUND( m_positionX ),
	                                                ROUND( m_positionY ) );

	LOG( 1 ) INFO( "Avatar::spawnAvatar: spawned at (%4i, %4i)\n", x, y );
	m_positionX = x;
	m_positionY = y;

	m_health = m_maxHealth;
	m_fuel   = m_maxFuel;
	setState( VISIBLE | WEAPON_VISIBLE | LIVING, true );

	// set avatar stencil
	setCollRect();
}

/**********************************************************/

Vector Avatar::placeBesideCollRect( const int marginLeft, const int marginRight,
		                                const int marginTop, const int marginBottom ) {
	const int dx = isFacingRight() ? m_collRectX + m_collRectWidth + marginRight :
	                                 m_collRectX - 1 - marginLeft;
	const int dy = (m_aimingAngle > 90.0) ? m_collRectY + m_collRectHeight + marginBottom :
	                                        m_collRectY - 1 - marginTop;
	const real tanValue = TAN_REAL( (isFacingRight() ? -1 : 1) * m_aimingAngle * AIMING_ANGLE_TO_RAD );

	Vector pos;

	if( FABS_REAL( tanValue ) > 1e-3 ) { // NOT almost vertical
		pos.y = dx / tanValue;
	}
	else pos.y = (m_aimingAngle > 90.0) ? 1e+10 : -1e+10;

	if( FABS_REAL( tanValue ) < 1e+3 ) { // NOT almost horizontal
		pos.x = dy * tanValue;
	}
	else pos.x = (isFacingRight() ? 1e+10 : -1e+10);

	pos.x = min( pos.x, static_cast<real>( m_collRectX + m_collRectWidth + marginRight ));
	pos.x = max( pos.x, static_cast<real>( m_collRectX - 1 - marginLeft ));
	pos.y = min( pos.y, static_cast<real>( m_collRectY + m_collRectHeight + marginBottom ));
	pos.y = max( pos.y, static_cast<real>( m_collRectY - 1 - marginTop ));

	return pos + getPosition();
}

/**********************************************************/

void Avatar::serialize( Uint8*& bufferPointer ) const
{
	// expands to a check of the buffer movement
	START_OBJECT_SERIALIZED_SIZE_CHECK( bufferPointer );

	// serialize base object
	CollidableObject::serialize( bufferPointer );	
	// Avatar specific serialisation
	Serialize<Uint32>::serialize( (int)getNumColors(),
	                              (Uint32*)getColors(),
	                               bufferPointer );
	Serialize<Sint32>::serialize( m_maxHealth   , bufferPointer );
	Serialize<Sint32>::serialize( m_fuel        , bufferPointer );
	Serialize<Sint32>::serialize( m_maxFuel     , bufferPointer );
	Serialize<real  >::serialize( m_aimingAngle , bufferPointer );
	Serialize<real  >::serialize( m_aimingSpeed , bufferPointer );
	Serialize<real  >::serialize( m_walkingSpeed, bufferPointer );

	// serialize the bonus pack
	m_bonusPack.serialize( bufferPointer );
	
	// expands to tag serialization
	SERIALIZE_OBJECT_TAG( bufferPointer );
	// expands to a check of the buffer movement
	END_OBJECT_SERIALIZED_SIZE_CHECK( bufferPointer, Avatar );
}

/**********************************************************/

void Avatar::deserialize( Uint8*& bufferPointer )
{
	// deserialize base object
	CollidableObject::deserialize( bufferPointer );
	// Avatar specific deserialisation
	Serialize<Uint32>::deserialize( (int)getNumColors(),
	                                 bufferPointer,
	                                (Uint32*)getColors() );
	Serialize<Sint32>::deserialize( bufferPointer, m_maxHealth    );
	Serialize<Sint32>::deserialize( bufferPointer, m_fuel         );
	Serialize<Sint32>::deserialize( bufferPointer, m_maxFuel      );
	Serialize<real  >::deserialize( bufferPointer, m_aimingAngle  );
	Serialize<real  >::deserialize( bufferPointer, m_aimingSpeed  );
	Serialize<real  >::deserialize( bufferPointer, m_walkingSpeed );

	// deserialize the bonus pack
	m_bonusPack.deserialize( bufferPointer );

	// expands to tag deserialization
	DESERIALIZE_OBJECT_TAG( bufferPointer );
}

/**********************************************************/

Uint32 Avatar::getSerializeBufferSize() const {
	return   CollidableObject::getSerializeBufferSize()   // base object
	       + getNumColors() * Serialize<Uint32>::sizeOf() // colors
	       + Serialize<Sint32>::sizeOf( m_maxHealth )
	       + Serialize<Sint32>::sizeOf( m_fuel         )
	       + Serialize<Sint32>::sizeOf( m_maxFuel      )
	       + Serialize<real  >::sizeOf( m_aimingAngle  )
	       + Serialize<real  >::sizeOf( m_aimingSpeed  )
	       + Serialize<real  >::sizeOf( m_walkingSpeed )
	       + m_bonusPack.getSerializeBufferSize()
	         // adds additional size for debugging (see serialize.hpp),
	         // but only, if the tags are used
	         PLUS_TAG_SIZE( 1 );
}

/**********************************************************/
