/* $Id: goal.cpp,v 1.2 2005/02/05 09:39:37 pohlt Exp $ */

#include "global.hpp"
#include "spriteset.hpp"
#include "world.hpp"
#include "goal.hpp"

/**********************************************************/

Goal::Goal() {
	setDrawBeforeCollidableObjects( true );
}

/**********************************************************/

Goal::~Goal() {
}

/**********************************************************/

bool Goal::initialize( const int posX, const int posY,
                       const Vector& velocity ) {
	m_positionX = posX;   m_positionY = posY;
	m_velocity = velocity;
	m_newDirDelay = m_initialNewDirDelay;

	return true;
}

/**********************************************************/

void Goal::update() {
	// move goal
	m_positionX += m_velocity.x;
	m_positionY += m_velocity.y;

	bool newVelocityOk = true;

	if( m_positionX-m_size < 0 ||
	    m_positionX+m_size > m_worldPointer->getMap()->getSizeX() ) {
		m_velocity.x *= -1;
		newVelocityOk = false;
	}
	if( m_positionY-m_size < 0 ||
	    m_positionY+m_size > m_worldPointer->getMap()->getSizeY() ) {
		m_velocity.y *= -1;
		newVelocityOk = false;
	}

	if( --m_newDirDelay <= 0 && newVelocityOk )	{
		m_velocity = globalRnd.getVector( m_velocity.abs() );
		m_newDirDelay = m_initialNewDirDelay;
	}
}
/**********************************************************/

bool Goal::testInside( const Vector& pos ) {
	return (pos-getPosition()).abs2() <= m_size*m_size;
}

/**********************************************************/

Uint32 Goal::getSerializeBufferSize() const {
	return NonCollidableObject::getSerializeBufferSize() +
	       sizeof( m_velocity ) +
	       sizeof( m_newDirDelay )
	       PLUS_TAG_SIZE( 1 );
}

/**********************************************************/

void Goal::serialize( Uint8*& bufferPointer ) const {
	// expands to a check of the buffer movement
	START_OBJECT_SERIALIZED_SIZE_CHECK( bufferPointer );
	
	NonCollidableObject::serialize( bufferPointer );

	Serialize<Vector>::serialize( m_velocity, bufferPointer );
	Serialize<Sint32>::serialize( m_newDirDelay, bufferPointer );

	// expands to tag serialization
	SERIALIZE_OBJECT_TAG( bufferPointer );
	// expands to a check of the buffer movement
	END_OBJECT_SERIALIZED_SIZE_CHECK( bufferPointer, Goal );
}

/**********************************************************/

void Goal::deserialize( Uint8*& bufferPointer ) {
	NonCollidableObject::deserialize( bufferPointer );

	Serialize<Vector>::deserialize( bufferPointer, m_velocity );
	Serialize<Sint32>::deserialize( bufferPointer, m_newDirDelay );

	// expands to tag deserialization
	INFO( "testing ball object tag\n" );
	DESERIALIZE_OBJECT_TAG( bufferPointer );
}

/**********************************************************/
